@@ -16,6 +16,7 @@ import threading
 import time
 from collections import defaultdict, namedtuple
 from typing import Callable, Optional, Union
+from urllib.parse import urlparse
 
 import attrdict
 import requests
@@ -25,6 +26,7 @@ from deprecated import deprecated
 from six.moves import urllib
 
 from . import requests_usbmux, xcui_element_types
+from ._proto import *
 from .exceptions import *
 from .utils import inject_call
 
@@ -33,18 +35,14 @@ try:
 except ImportError:
     from cached_property import cached_property
 
-
-logger = logging.getLogger("facebook-wda")
-urlparse = urllib.parse.urlparse
-_urljoin = urllib.parse.urljoin
-_is_tmq = os.getenv("TMQ") == "true"
-
-if six.PY3:
-    from functools import reduce
+try:
+    from logzero import logger
+except ImportError:
+    logger = logging.getLogger("facebook-wda")  # default level: WARNING
 
 DEBUG = False
 HTTP_TIMEOUT = 60.0  # unit second
-DEVICE_WAIT_TIMEOUT = 30.0 # wait ready
+DEVICE_WAIT_TIMEOUT = 120.0  # wait ready
 
 LANDSCAPE = 'LANDSCAPE'
 PORTRAIT = 'PORTRAIT'
@@ -125,6 +123,22 @@ def _requests_session_pool_get(scheme, netloc):
     return requests_usbmux.Session()
 
 
+def _is_tmq_platform() -> bool:
+    return os.getenv("TMQ") == "true"
+
+
+def _is_session_id_error(value: Union[str, dict]):
+    if isinstance(value, dict):  # 新版WDA逻辑
+        error = value.get('error')
+        if error == "invalid session id" or "possibly crashed" in value.get(
+                'message', ''):
+            return True
+    elif isinstance(value, str):  # 旧版WDA中的value就是字符串
+        if "invalid session id" in value or "possibly crashed" in value:
+            return True
+    return False
+
+
 def _unsafe_httpdo(url, method='GET', data=None):
     """
     Do HTTP Request
@@ -146,6 +160,9 @@ def _unsafe_httpdo(url, method='GET', data=None):
             requests.exceptions.ReadTimeout) as e:
         raise
 
+    if response.status_code == 502: # Bad Gateway
+        raise WDABadGateway(response.status_code, response.text)
+
     if DEBUG:
         ms = (time.time() - start) * 1000
         print('Return ({:.0f}ms): {}'.format(ms, response.text))
@@ -154,18 +171,20 @@ def _unsafe_httpdo(url, method='GET', data=None):
         retjson = response.json()
         retjson['status'] = retjson.get('status', 0)
         r = convert(retjson)
-        if r.status != 0:
+
+        if r.status != 0:  # in new WDA, status is always 0
+            if _is_session_id_error(r.value):
+                r['status'] = Status.INVALID_SESSION_ID
             raise WDARequestError(r.status, r.value)
+
         if isinstance(r.value, dict) and r.value.get("error"):
-            error = r.value['error']
-            if error == "invalid session id" or "possibly crashed" in r.value.get(
-                    'message', ''):
+            if _is_session_id_error(r.value):
                 status = Status.INVALID_SESSION_ID
             else:
                 status = Status.ERROR
-            # value = r.value.copy()
-            # value.pop("traceback", None)
-            raise WDARequestError(status, r.value)
+            value = r.value.copy()
+            value.pop("traceback", None)
+            raise WDARequestError(status, value)
         return r
     except JSONDecodeError:
         if response.text == "":
@@ -243,14 +262,16 @@ class BaseClient(object):
     def _callback_fix_invalid_session_id(self, err: WDAError):
         """ 当遇到 invalid session id错误时，更新session id并重试 """
         if isinstance(err, WDARequestError) and \
-                err.status == Status.INVALID_SESSION_ID and err.value.error == "invalid session id":
+                err.status == Status.INVALID_SESSION_ID: # and not self.__is_app:
+            self.close()
             self.session_id = None
             return Callback.RET_RETRY
 
-    def _callback_wait_ready(self, err):
         """ 等待设备恢复上线 """
-        if isinstance(err, (ConnectionError, requests.ConnectionError)):
-            if not self.wait_ready(DEVICE_WAIT_TIMEOUT): # 等待设备恢复在线
+    def _callback_wait_ready(self, err):
+        logger.warning("Error: %s", err)
+        if isinstance(err, (ConnectionError, requests.ConnectionError, requests.ReadTimeout, WDABadGateway)):
+            if not self.wait_ready(DEVICE_WAIT_TIMEOUT):  # 等待设备恢复在线
                 return Callback.RET_ABORT
             return Callback.RET_RETRY
 
@@ -261,22 +282,27 @@ class BaseClient(object):
             print("send_keys callback called")
 
     def _callback_tmq_print_error(self, method, url, data, err):
-        logger.warning("HTTP Error happens, this message is printed for better debugging")
+        logger.warning(
+            "HTTP Error happens, this message is printed for better debugging")
         body = json.dumps(data) if data else ''
         logger.warning("Shell$ curl -X {method} -d '{body}' '{url}'".format(
             method=method.upper(), body=body or '', url=url))
         logger.warning("Error: %s", err)
 
     def _init_fix(self):
-        self.register_callback(Callback.ERROR, self._callback_fix_invalid_session_id)
-        self.register_callback(Callback.ERROR, self._callback_wait_ready)
-        if _is_tmq:
+        self.register_callback(Callback.ERROR,
+                               self._callback_fix_invalid_session_id)
+        if _is_tmq_platform():
             # 输入之前处理弹窗
             # 出现错误是print出来，方便调试
-            self.register_callback(Callback.HTTP_REQUEST_BEFORE, self._callback_tmq_before_send_keys)
-            self.register_callback(Callback.ERROR, self._callback_tmq_print_error)
+            logger.info("register callbacks for tmq")
+            self.register_callback(Callback.ERROR, self._callback_wait_ready)
+            self.register_callback(Callback.HTTP_REQUEST_BEFORE,
+                                   self._callback_tmq_before_send_keys)
+            self.register_callback(Callback.ERROR,
+                                   self._callback_tmq_print_error)
 
-    def wait_ready(self, timeout=120, noprint = False) -> bool:
+    def wait_ready(self, timeout=120, noprint=False) -> bool:
         """
         wait until WDA back to normal
 
@@ -284,11 +310,12 @@ class BaseClient(object):
             bool (if wda works)
         """
         deadline = time.time() + timeout
+
         def _dprint(message: str):
             if noprint:
                 return
             print(message)
-        
+
         _dprint("Wait ready (timeout={:.1f})".format(timeout))
         while time.time() < deadline:
             try:
@@ -296,7 +323,8 @@ class BaseClient(object):
                 _dprint("device back online")
                 return True
             except:
-                _dprint("wait_ready left {:.1f} seconds".format(deadline - time.time()))
+                _dprint("wait_ready left {:.1f} seconds".format(deadline -
+                                                                time.time()))
                 time.sleep(1.0)
         _dprint("device still offline")
         return False
@@ -335,13 +363,17 @@ class BaseClient(object):
             ]:
                 return ret
 
+    @property
+    def callbacks(self):
+        return self.__callbacks
+
     def _fetch(self,
                method: str,
                urlpath: str,
                data: Optional[dict] = None,
                with_session: bool = False) -> attrdict.AttrDict:
         """ do http request """
-        urlpath = "/" + urlpath.lstrip("/") # urlpath always startswith /
+        urlpath = "/" + urlpath.lstrip("/")  # urlpath always startswith /
 
         callbacks = None
         if self.__callback_depth == 0:
@@ -360,13 +392,13 @@ class BaseClient(object):
 
         try:
             if with_session:
-                url = urljoin(self.__wda_url, "session", self.session_id, urlpath)
+                url = urljoin(self.__wda_url, "session", self.session_id,
+                              urlpath)
             run_callback(Callback.HTTP_REQUEST_BEFORE)
             response = httpdo(url, method, data)
             run_callback(Callback.HTTP_REQUEST_AFTER, response=response)
             return response
-        except (WDARequestError, ConnectionError,
-                requests.ConnectionError) as err:
+        except Exception as err:
             ret = run_callback(Callback.ERROR, err=err)
             if ret == Callback.RET_RETRY:
                 return self._fetch(method, urlpath, data, with_session)
@@ -379,16 +411,18 @@ class BaseClient(object):
 
     @property
     def http(self):
-        return namedtuple("HTTPRequest", ['get', 'post'])(
+        return namedtuple("HTTPRequest", ['fetch', 'get', 'post'])(
+            self._fetch,
             functools.partial(self._fetch, "GET"),
             functools.partial(self._fetch, "POST")) # yapf: disable
 
     @property
     def _session_http(self):
-        return namedtuple("HTTPSessionRequest", ['fetch', 'get', 'post'])(
+        return namedtuple("HTTPSessionRequest", ['fetch', 'get', 'post', 'delete'])(
             functools.partial(self._fetch, with_session=True),
             functools.partial(self._fetch, "GET", with_session=True),
-            functools.partial(self._fetch, "POST", with_session=True)) # yapf: disable
+            functools.partial(self._fetch, "POST", with_session=True),
+            functools.partial(self._fetch, "DELETE", with_session=True)) # yapf: disable
 
     def home(self):
         """Press home button"""
@@ -471,7 +505,7 @@ class BaseClient(object):
                 bundle_id=None,
                 arguments: Optional[list] = None,
                 environment: Optional[dict] = None,
-                alert_action: Optional[str] = None):
+                alert_action: Optional[AlertAction] = None):
         """
         Launch app in a session
 
@@ -479,7 +513,7 @@ class BaseClient(object):
             - bundle_id (str): the app bundle id
             - arguments (list): ['-u', 'https://www.google.com/ncr']
             - enviroment (dict): {"KEY": "VAL"}
-            - alert_action (str): "accept" or "dismiss"
+            - alert_action (AlertAction): AlertAction.ACCEPT or AlertAction.DISMISS
 
         WDA Return json like
 
@@ -554,7 +588,11 @@ class BaseClient(object):
         return Client(self.__wda_url, _session_id=res.sessionId)
 
     def close(self):  # close session
-        return self._session_http.delete('/')
+        try:
+            return self._session_http.delete('/')
+        except WDARequestError as e:
+            if e.status != Status.INVALID_SESSION_ID:
+                raise
 
     #@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@#
     ######  Session methods and properties ######
@@ -771,7 +809,7 @@ class BaseClient(object):
                                        dict(duration=duration))
 
     def tap(self, x, y):
-        #if _is_tmq:
+        #if _is_tmq_platform():
         #    return self.http.post("/mds/touchAndHold", dict(x=x, y=y, duration=0.02))
         return self._session_http.post('/wda/tap/0', dict(x=x, y=y))
 
@@ -910,7 +948,12 @@ class BaseClient(object):
         """
         Get and set /session/$sessionId/appium/settings
         """
-        return self._session_http.get("/appium/settings").value
+        if value is None:
+            return self._session_http.get("/appium/settings").value
+        return self._session_http.post("/appium/settings",
+                                       data={
+                                           "settings": value
+                                       }).value
 
     def xpath(self, value):
         """
@@ -1247,7 +1290,7 @@ class Selector(object):
                 return elems[0]
             if start_time + timeout < time.time():
                 break
-            time.sleep(0.01)
+            time.sleep(0.5)
 
         if raise_error:
             raise WDAElementNotFoundError("element not found",
@@ -1256,9 +1299,10 @@ class Selector(object):
     def __getattr__(self, oper):
         if oper.startswith("_"):
             raise AttributeError("invalid attr", oper)
+        if not hasattr(Element, oper):
+            raise AttributeError("'Element' object has no attribute %r" % oper)
+
         el = self.get()
-        if not hasattr(el, oper):
-            raise AttributeError("invalid attr", oper)
         return getattr(el, oper)
 
     def set_timeout(self, s):
@@ -1281,6 +1325,16 @@ class Selector(object):
     def exists(self):
         return len(self.find_element_ids()) > self._index
 
+    def click(self, timeout: Optional[float] = None):
+        """
+        Click element
+
+        Args:
+            timeout (float): max wait seconds
+        """
+        e = self.get(timeout=timeout)
+        e.click()
+
     def click_exists(self, timeout=0):
         """
         Wait element and perform click
@@ -1369,6 +1423,22 @@ class Element(object):
         ret = self.http.get('/wda/element/%s/%s' % (self._id, key)).value
         return ret
 
+    @property
+    def info(self):
+        return {
+            "id": self.id,
+            "label": self.label,
+            "value": self.value,
+            "text": self.text,
+            "name": self.name,
+            "className": self.className,
+            "enabled": self.enabled,
+            "displayed": self.displayed,
+            "visible": self.visible,
+            "accessible": self.accessible,
+            "accessibilityContainer": self.accessibility_container
+        }
+
     @property
     def id(self):
         return self._id
@@ -1409,10 +1479,6 @@ class Element(object):
     def value(self):
         return self._prop('attribute/value')
 
-    @property
-    def enabled(self):
-        return self._prop('enabled')
-
     @property
     def visible(self):
         return self._prop('attribute/visible')
@@ -1428,7 +1494,7 @@ class Element(object):
     def tap(self):
         return self._req('post', '/click')
 
-    def click(self, scroll: str = None):
+    def click(self):
         """
         Get element center position and do click, a little slower
         """
